@@ -3,6 +3,7 @@
 #
 # License: BSD (3-clause)
 
+from contextlib import contextmanager
 from distutils.version import LooseVersion
 import gc
 import os
@@ -12,29 +13,21 @@ import shutil
 import sys
 import warnings
 import pytest
-# For some unknown reason, on Travis-xenial there are segfaults caused on
-# the line pytest -> pdb.Pdb.__init__ -> "import readline". Forcing an
-# import here seems to prevent them (!?). This suggests a potential problem
-# with some other library stepping on memory where it shouldn't. It only
-# seems to happen on the Linux runs that install Mayavi. Anectodally,
-# @larsoner has had problems a couple of years ago where a mayavi import
-# seemed to corrupt SciPy linalg function results (!), likely due to the
-# associated VTK import, so this could be another manifestation of that.
-try:
-    import readline  # noqa
-except Exception:
-    pass
 
 import numpy as np
+
 import mne
 from mne.datasets import testing
-from mne.utils import _pl
+from mne.fixes import has_numba
+from mne.stats import cluster_level
+from mne.utils import _pl, _assert_no_instances, numerics
 
 test_path = testing.data_path(download=False)
 s_path = op.join(test_path, 'MEG', 'sample')
 fname_evoked = op.join(s_path, 'sample_audvis_trunc-ave.fif')
 fname_cov = op.join(s_path, 'sample_audvis_trunc-cov.fif')
 fname_fwd = op.join(s_path, 'sample_audvis_trunc-meg-eeg-oct-4-fwd.fif')
+fname_fwd_full = op.join(s_path, 'sample_audvis_trunc-meg-eeg-oct-6-fwd.fif')
 bem_path = op.join(test_path, 'subjects', 'sample', 'bem')
 fname_bem = op.join(bem_path, 'sample-1280-bem.fif')
 fname_aseg = op.join(test_path, 'subjects', 'sample', 'mri', 'aseg.mgz')
@@ -62,6 +55,7 @@ def pytest_configure(config):
     #   doc/conf.py.
     warning_lines = r"""
     error::
+    ignore:.*deprecated and ignored since IPython.*:DeprecationWarning
     ignore::ImportWarning
     ignore:the matrix subclass:PendingDeprecationWarning
     ignore:numpy.dtype size changed:RuntimeWarning
@@ -101,6 +95,9 @@ def pytest_configure(config):
     ignore:.*tostring.*is deprecated.*:DeprecationWarning
     ignore:.*QDesktopWidget\.availableGeometry.*:DeprecationWarning
     ignore:Unable to enable faulthandler.*:UserWarning
+    ignore:Fetchers from the nilearn.*:FutureWarning
+    ignore:SelectableGroups dict interface is deprecated\. Use select\.:DeprecationWarning
+    ignore:Call to deprecated class vtk.*:DeprecationWarning
     always:.*get_data.* is deprecated in favor of.*:DeprecationWarning
     always::ResourceWarning
     """  # noqa: E501
@@ -126,6 +123,15 @@ def check_verbose(request):
                     ' modifies logger.level')
 
 
+@pytest.fixture(autouse=True)
+def close_all():
+    """Close all matplotlib plots, regardless of test status."""
+    # This adds < 1 µS in local testing, and we have ~2500 tests, so ~2 ms max
+    import matplotlib.pyplot as plt
+    yield
+    plt.close('all')
+
+
 @pytest.fixture(scope='function')
 def verbose_debug():
     """Run a test with debug verbosity."""
@@ -230,7 +236,8 @@ def bias_params_free(evoked, noise_cov):
 def bias_params_fixed(evoked, noise_cov):
     """Provide inputs for fixed bias functions."""
     fwd = mne.read_forward_solution(fname_fwd)
-    fwd = mne.convert_forward_solution(fwd, force_fixed=True, surf_ori=True)
+    mne.convert_forward_solution(
+        fwd, force_fixed=True, surf_ori=True, copy=False)
     return _bias_params(evoked, noise_cov, fwd)
 
 
@@ -238,14 +245,23 @@ def _bias_params(evoked, noise_cov, fwd):
     evoked.pick_types(meg=True, eeg=True, exclude=())
     # restrict to limited set of verts (small src here) and one hemi for speed
     vertices = [fwd['src'][0]['vertno'].copy(), []]
-    stc = mne.SourceEstimate(np.zeros((sum(len(v) for v in vertices), 1)),
-                             vertices, 0., 1.)
+    stc = mne.SourceEstimate(
+        np.zeros((sum(len(v) for v in vertices), 1)), vertices, 0, 1)
     fwd = mne.forward.restrict_forward_to_stc(fwd, stc)
     assert fwd['sol']['row_names'] == noise_cov['names']
     assert noise_cov['names'] == evoked.ch_names
     evoked = mne.EvokedArray(fwd['sol']['data'].copy(), evoked.info)
     data_cov = noise_cov.copy()
-    data_cov['data'] = np.dot(fwd['sol']['data'], fwd['sol']['data'].T)
+    data = fwd['sol']['data'] @ fwd['sol']['data'].T
+    data *= 1e-14  # 100 nAm at each source, effectively (1e-18 would be 1 nAm)
+    # This is rank-deficient, so let's make it actually positive semidefinite
+    # by regularizing a tiny bit
+    data.flat[::data.shape[0] + 1] += mne.make_ad_hoc_cov(evoked.info)['data']
+    # Do our projection
+    proj, _, _ = mne.io.proj.make_projector(
+        data_cov['projs'], data_cov['names'])
+    data = proj @ data @ proj.T
+    data_cov['data'][:] = data
     assert data_cov['data'].shape[0] == len(noise_cov['names'])
     want = np.arange(fwd['sol']['data'].shape[1])
     if not mne.forward.is_fixed_orient(fwd):
@@ -253,76 +269,83 @@ def _bias_params(evoked, noise_cov, fwd):
     return evoked, fwd, noise_cov, data_cov, want
 
 
-@pytest.fixture(scope="module", params=[
-    "mayavi",
-    "pyvista",
-])
-def backend_name(request):
-    """Get the backend name."""
-    yield request.param
-
-
-@pytest.yield_fixture
-def renderer(backend_name, garbage_collect):
-    """Yield the 3D backends."""
-    from mne.viz.backends.renderer import _use_test_3d_backend
-    _check_skip_backend(backend_name)
-    with _use_test_3d_backend(backend_name):
-        from mne.viz.backends import renderer
-        yield renderer
-        renderer.backend._close_all()
-
-
-@pytest.yield_fixture
+@pytest.fixture
 def garbage_collect():
     """Garbage collect on exit."""
     yield
     gc.collect()
 
 
-@pytest.fixture(scope="module", params=[
-    "pyvista",
-    "mayavi",
-])
-def backend_name_interactive(request):
-    """Get the backend name."""
-    yield request.param
-
-
-@pytest.yield_fixture
-def renderer_interactive(backend_name_interactive):
+@pytest.fixture(params=["mayavi", "pyvista"])
+def renderer(request, garbage_collect):
     """Yield the 3D backends."""
-    from mne.viz.backends.renderer import _use_test_3d_backend
-    _check_skip_backend(backend_name_interactive)
-    with _use_test_3d_backend(backend_name_interactive, interactive=True):
-        from mne.viz.backends import renderer
+    with _use_backend(request.param, interactive=False) as renderer:
         yield renderer
-        renderer.backend._close_all()
+
+
+@pytest.fixture(params=["pyvista"])
+def renderer_pyvista(request, garbage_collect):
+    """Yield the PyVista backend."""
+    with _use_backend(request.param, interactive=False) as renderer:
+        yield renderer
+
+
+@pytest.fixture(params=["notebook"])
+def renderer_notebook(request):
+    """Yield the 3D notebook renderer."""
+    with _use_backend(request.param, interactive=False) as renderer:
+        yield renderer
+
+
+@pytest.fixture(scope="module", params=["pyvista"])
+def renderer_interactive_pyvista(request):
+    """Yield the interactive PyVista backend."""
+    with _use_backend(request.param, interactive=True) as renderer:
+        yield renderer
+
+
+@pytest.fixture(scope="module", params=["pyvista", "mayavi"])
+def renderer_interactive(request):
+    """Yield the interactive 3D backends."""
+    with _use_backend(request.param, interactive=True) as renderer:
+        if renderer._get_3d_backend() == 'mayavi':
+            with warnings.catch_warnings(record=True):
+                try:
+                    from surfer import Brain  # noqa: 401 analysis:ignore
+                except Exception:
+                    pytest.skip('Requires PySurfer')
+        yield renderer
+
+
+@contextmanager
+def _use_backend(backend_name, interactive):
+    from mne.viz.backends.renderer import _use_test_3d_backend
+    _check_skip_backend(backend_name)
+    with _use_test_3d_backend(backend_name, interactive=interactive):
+        from mne.viz.backends import renderer
+        try:
+            yield renderer
+        finally:
+            renderer.backend._close_all()
 
 
 def _check_skip_backend(name):
     from mne.viz.backends.tests._utils import (has_mayavi, has_pyvista,
                                                has_pyqt5, has_imageio_ffmpeg)
+    check_pyvista = name in ('pyvista', 'notebook')
+    check_pyqt5 = name in ('mayavi', 'pyvista')
     if name == 'mayavi':
         if not has_mayavi():
             pytest.skip("Test skipped, requires mayavi.")
     elif name == 'pyvista':
-        if not has_pyvista():
-            pytest.skip("Test skipped, requires pyvista.")
         if not has_imageio_ffmpeg():
             pytest.skip("Test skipped, requires imageio-ffmpeg")
-    if not has_pyqt5():
+    if check_pyvista and not has_pyvista():
+        pytest.skip("Test skipped, requires pyvista.")
+    if check_pyqt5 and not has_pyqt5():
         pytest.skip("Test skipped, requires PyQt5.")
 
 
-@pytest.fixture()
-def renderer_notebook():
-    """Verify that pytest_notebook is installed."""
-    from mne.viz.backends.renderer import _use_test_3d_backend
-    with _use_test_3d_backend('notebook'):
-        yield
-
-
 @pytest.fixture(scope='session')
 def pixel_ratio():
     """Get the pixel ratio."""
@@ -473,11 +496,64 @@ def download_is_error(monkeypatch):
     monkeypatch.setattr(mne.utils.fetching, '_get_http', _fail)
 
 
+@pytest.fixture()
+def brain_gc(request):
+    """Ensure that brain can be properly garbage collected."""
+    keys = (
+        'renderer_interactive',
+        'renderer_interactive_pyvista',
+        'renderer_interactive_pysurfer',
+        'renderer',
+        'renderer_pyvista',
+        'renderer_notebook',
+    )
+    assert set(request.fixturenames) & set(keys) != set()
+    for key in keys:
+        if key in request.fixturenames:
+            is_pv = request.getfixturevalue(key)._get_3d_backend() == 'pyvista'
+            close_func = request.getfixturevalue(key).backend._close_all
+            break
+    if not is_pv:
+        yield
+        return
+    import pyvista
+    if LooseVersion(pyvista.__version__) <= LooseVersion('0.26.1'):
+        yield
+        return
+    from mne.viz import Brain
+    ignore = set(id(o) for o in gc.get_objects())
+    yield
+    close_func()
+    # no need to warn if the test itself failed, pytest-harvest helps us here
+    try:
+        outcome = request.node.harvest_rep_call
+    except Exception:
+        outcome = 'failed'
+    if outcome != 'passed':
+        return
+    _assert_no_instances(Brain, 'after')
+    # We only check VTK for PyVista -- Mayavi/PySurfer is not as strict
+    objs = gc.get_objects()
+    bad = list()
+    for o in objs:
+        try:
+            name = o.__class__.__name__
+        except Exception:  # old Python, probably
+            pass
+        else:
+            if name.startswith('vtk') and id(o) not in ignore:
+                bad.append(name)
+        del o
+    del objs, ignore, Brain
+    assert len(bad) == 0, 'VTK objects linger:\n' + '\n'.join(bad)
+
+
 def pytest_sessionfinish(session, exitstatus):
     """Handle the end of the session."""
     n = session.config.option.durations
     if n is None:
         return
+    print('\n')
     try:
         import pytest_harvest
     except ImportError:
@@ -510,3 +586,21 @@ def pytest_sessionfinish(session, exitstatus):
         timings = [timing.rjust(rjust) for timing in timings]
         for name, timing in zip(names, timings):
             writer.line(f'{timing.ljust(15)}{name}')
+
+
+@pytest.fixture(scope="function", params=('Numba', 'NumPy'))
+def numba_conditional(monkeypatch, request):
+    """Test both code paths on machines that have Numba."""
+    assert request.param in ('Numba', 'NumPy')
+    if request.param == 'NumPy' and has_numba:
+        monkeypatch.setattr(
+            cluster_level, '_get_buddies', cluster_level._get_buddies_fallback)
+        monkeypatch.setattr(
+            cluster_level, '_get_selves', cluster_level._get_selves_fallback)
+        monkeypatch.setattr(
+            cluster_level, '_where_first', cluster_level._where_first_fallback)
+        monkeypatch.setattr(
+            numerics, '_arange_div', numerics._arange_div_fallback)
+    if request.param == 'Numba' and not has_numba:
+        pytest.skip('Numba not installed')
+    yield request.param
