@@ -12,22 +12,35 @@ at which the fix is no longer needed.
 #          Lars Buitinck <L.J.Buitinck@uva.nl>
 # License: BSD
 
-import inspect
 from distutils.version import LooseVersion
+import functools
+import inspect
 from math import log
 import os
 from pathlib import Path
 import warnings
 
 import numpy as np
-import scipy
-from scipy import linalg
-from scipy.linalg import LinAlgError
 
 
 ###############################################################################
 # Misc
 
+def _median_complex(data, axis):
+    """Compute marginal median on complex data safely.
+
+    XXX: Can be removed when numpy introduces a fix.
+    See: https://github.com/scipy/scipy/pull/12676/.
+    """
+    # np.median must be passed real arrays for the desired result
+    if np.iscomplexobj(data):
+        data = (np.median(np.real(data), axis=axis)
+                + 1j * np.median(np.imag(data), axis=axis))
+    else:
+        data = np.median(data, axis=axis)
+    return data
+
+
 # helpers to get function arguments
 def _get_args(function, varargs=False):
     params = inspect.signature(function).parameters
@@ -49,6 +62,7 @@ def _safe_svd(A, **kwargs):
     #     https://software.intel.com/en-us/forums/intel-distribution-for-python/topic/628049  # noqa: E501
     # For SciPy 0.18 and up, we can work around it by using
     # lapack_driver='gesvd' instead.
+    from scipy import linalg
     if kwargs.get('overwrite_a', False):
         raise ValueError('Cannot set overwrite_a=True with this function')
     try:
@@ -63,6 +77,11 @@ def _safe_svd(A, **kwargs):
             raise
 
 
+def _csc_matrix_cast(x):
+    from scipy.sparse import csc_matrix
+    return csc_matrix(x)
+
+
 ###############################################################################
 # Backporting nibabel's read_geometry
 
@@ -73,7 +92,7 @@ def _get_read_geometry():
         has_nibabel = True
     except ImportError:
         has_nibabel = False
-    if has_nibabel and LooseVersion(nib.__version__) > LooseVersion('2.1.0'):
+    if has_nibabel:
         from nibabel.freesurfer import read_geometry
     else:
         read_geometry = _read_geometry
@@ -141,196 +160,27 @@ def _read_geometry(filepath, read_metadata=False, read_stamp=False):
     return ret
 
 
-###############################################################################
-# Triaging scipy.signal.windows.dpss (1.1)
-
-def tridisolve(d, e, b, overwrite_b=True):
-    """Symmetric tridiagonal system solver, from Golub and Van Loan p157.
-
-    .. note:: Copied from NiTime.
-
-    Parameters
-    ----------
-    d : ndarray
-      main diagonal stored in d[:]
-    e : ndarray
-      superdiagonal stored in e[:-1]
-    b : ndarray
-      RHS vector
-
-    Returns
-    -------
-    x : ndarray
-      Solution to Ax = b (if overwrite_b is False). Otherwise solution is
-      stored in previous RHS vector b
-
-    """
-    N = len(b)
-    # work vectors
-    dw = d.copy()
-    ew = e.copy()
-    if overwrite_b:
-        x = b
-    else:
-        x = b.copy()
-    for k in range(1, N):
-        # e^(k-1) = e(k-1) / d(k-1)
-        # d(k) = d(k) - e^(k-1)e(k-1) / d(k-1)
-        t = ew[k - 1]
-        ew[k - 1] = t / dw[k - 1]
-        dw[k] = dw[k] - t * ew[k - 1]
-    # This iterative solver can fail sometimes. There is probably a
-    # graceful way to solve this, but it should only be a problem
-    # in very rare cases. Users of SciPy 1.1+ will never hit this anyway,
-    # so not worth spending more time figuring out how to do it faster.
-    if dw[N - 1] == 0:
-        a = np.diag(d) + np.diag(e[:-1], -1) + np.diag(e[:-1], 1)
-        x[:] = linalg.solve(a, b)
-    else:
-        for k in range(1, N):
-            x[k] = x[k] - ew[k - 1] * x[k - 1]
-        if dw[N - 1] != 0:
-            x[N - 1] = x[N - 1] / dw[N - 1]
-        for k in range(N - 2, -1, -1):
-            x[k] = x[k] / dw[k] - ew[k] * x[k + 1]
-
-    if not overwrite_b:
-        return x
-
-
-def tridi_inverse_iteration(d, e, w, x0=None, rtol=1e-8):
-    """Perform an inverse iteration.
-
-    This will find the eigenvector corresponding to the given eigenvalue
-    in a symmetric tridiagonal system.
-
-    ..note:: Copied from NiTime.
-
-    Parameters
-    ----------
-    d : ndarray
-      main diagonal of the tridiagonal system
-    e : ndarray
-      offdiagonal stored in e[:-1]
-    w : float
-      eigenvalue of the eigenvector
-    x0 : ndarray
-      initial point to start the iteration
-    rtol : float
-      tolerance for the norm of the difference of iterates
-
-    Returns
-    -------
-    e: ndarray
-      The converged eigenvector
-    """
-    eig_diag = d - w
-    if x0 is None:
-        x0 = np.random.randn(len(d))
-    x_prev = np.zeros_like(x0)
-    norm_x = np.linalg.norm(x0)
-    # the eigenvector is unique up to sign change, so iterate
-    # until || |x^(n)| - |x^(n-1)| ||^2 < rtol
-    x0 /= norm_x
-    while np.linalg.norm(np.abs(x0) - np.abs(x_prev)) > rtol:
-        x_prev = x0.copy()
-        tridisolve(eig_diag, e, x0)
-        norm_x = np.linalg.norm(x0)
-        x0 /= norm_x
-    return x0
-
-
-def _dpss(N, half_nbw, Kmax):
-    """Compute DPSS windows."""
-    # here we want to set up an optimization problem to find a sequence
-    # whose energy is maximally concentrated within band [-W,W].
-    # Thus, the measure lambda(T,W) is the ratio between the energy within
-    # that band, and the total energy. This leads to the eigen-system
-    # (A - (l1)I)v = 0, where the eigenvector corresponding to the largest
-    # eigenvalue is the sequence with maximally concentrated energy. The
-    # collection of eigenvectors of this system are called Slepian
-    # sequences, or discrete prolate spheroidal sequences (DPSS). Only the
-    # first K, K = 2NW/dt orders of DPSS will exhibit good spectral
-    # concentration
-    # [see http://en.wikipedia.org/wiki/Spectral_concentration_problem]
-
-    # Here I set up an alternative symmetric tri-diagonal eigenvalue
-    # problem such that
-    # (B - (l2)I)v = 0, and v are our DPSS (but eigenvalues l2 != l1)
-    # the main diagonal = ([N-1-2*t]/2)**2 cos(2PIW), t=[0,1,2,...,N-1]
-    # and the first off-diagonal = t(N-t)/2, t=[1,2,...,N-1]
-    # [see Percival and Walden, 1993]
-    nidx = np.arange(N, dtype='d')
-    W = float(half_nbw) / N
-    diagonal = ((N - 1 - 2 * nidx) / 2.) ** 2 * np.cos(2 * np.pi * W)
-    off_diag = np.zeros_like(nidx)
-    off_diag[:-1] = nidx[1:] * (N - nidx[1:]) / 2.
-    # put the diagonals in LAPACK "packed" storage
-    ab = np.zeros((2, N), 'd')
-    ab[1] = diagonal
-    ab[0, 1:] = off_diag[:-1]
-    # only calculate the highest Kmax eigenvalues
-    w = linalg.eigvals_banded(ab, select='i',
-                              select_range=(N - Kmax, N - 1))
-    w = w[::-1]
-
-    # find the corresponding eigenvectors via inverse iteration
-    t = np.linspace(0, np.pi, N)
-    dpss = np.zeros((Kmax, N), 'd')
-    for k in range(Kmax):
-        dpss[k] = tridi_inverse_iteration(diagonal, off_diag, w[k],
-                                          x0=np.sin((k + 1) * t))
-
-    # By convention (Percival and Walden, 1993 pg 379)
-    # * symmetric tapers (k=0,2,4,...) should have a positive average.
-    # * antisymmetric tapers should begin with a positive lobe
-    fix_symmetric = (dpss[0::2].sum(axis=1) < 0)
-    for i, f in enumerate(fix_symmetric):
-        if f:
-            dpss[2 * i] *= -1
-    # rather than test the sign of one point, test the sign of the
-    # linear slope up to the first (largest) peak
-    pk = np.argmax(np.abs(dpss[1::2, :N // 2]), axis=1)
-    for i, p in enumerate(pk):
-        if np.sum(dpss[2 * i + 1, :p]) < 0:
-            dpss[2 * i + 1] *= -1
-
-    return dpss
-
-
-def _get_dpss():
-    try:
-        from scipy.signal.windows import dpss
-    except ImportError:
-        dpss = _dpss
-    return dpss
-
-
 ###############################################################################
 # Triaging FFT functions to get fast pocketfft (SciPy 1.4)
 
-try:
-    from scipy.fft import fft, ifft, fftfreq, rfft, irfft, rfftfreq, ifftshift
-except ImportError:
-    from numpy.fft import fft, ifft, fftfreq, rfft, irfft, rfftfreq, ifftshift
+@functools.lru_cache(None)
+def _import_fft(name):
+    single = False
+    if not isinstance(name, tuple):
+        name = (name,)
+        single = True
+    try:
+        from scipy.fft import rfft  # noqa analysis:ignore
+    except ImportError:
+        from numpy import fft  # noqa
+    else:
+        from scipy import fft  # noqa
+    out = [getattr(fft, n) for n in name]
+    if single:
+        out = out[0]
+    return out
 
 
-###############################################################################
-# Orth with rcond argument (SciPy 1.1)
-
-if LooseVersion(scipy.__version__) >= '1.1':
-    from scipy.linalg import orth
-else:
-    def orth(A, rcond=None):  # noqa
-        u, s, vh = linalg.svd(A, full_matrices=False)
-        M, N = u.shape[0], vh.shape[1]
-        if rcond is None:
-            rcond = np.finfo(s.dtype).eps * max(M, N)
-        tol = np.amax(s) * rcond
-        num = np.sum(s > tol, dtype=int)
-        Q = u[:, :num]
-        return Q
-
 ###############################################################################
 # NumPy Generator (NumPy 1.17)
 
@@ -357,7 +207,8 @@ def _validate_sos(sos):
 ###############################################################################
 # Misc utilities
 
-# Deal with nibabel 2.5 img.get_data() deprecation
+# get_fdata() requires knowing the dtype ahead of time, so let's triage on our
+# own instead
 def _get_img_fdata(img):
     data = np.asanyarray(img.dataobj)
     dtype = np.complex128 if np.iscomplexobj(data) else np.float64
@@ -458,8 +309,30 @@ def is_regressor(estimator):
     return getattr(estimator, "_estimator_type", None) == "regressor"
 
 
+_DEFAULT_TAGS = {
+    'non_deterministic': False,
+    'requires_positive_X': False,
+    'requires_positive_y': False,
+    'X_types': ['2darray'],
+    'poor_score': False,
+    'no_validation': False,
+    'multioutput': False,
+    "allow_nan": False,
+    'stateless': False,
+    'multilabel': False,
+    '_skip_test': False,
+    '_xfail_checks': False,
+    'multioutput_only': False,
+    'binary_only': False,
+    'requires_fit': True,
+    'preserves_dtype': [np.float64],
+    'requires_y': False,
+    'pairwise': False,
+}
+
+
 class BaseEstimator(object):
-    """Base class for all estimators in scikit-learn
+    """Base class for all estimators in scikit-learn.
 
     Notes
     -----
@@ -500,13 +373,13 @@ class BaseEstimator(object):
 
         Parameters
         ----------
-        deep : boolean, optional
+        deep : bool, optional
             If True, will return the parameters for this estimator and
             contained subobjects that are estimators.
 
         Returns
         -------
-        params : mapping of string to any
+        params : dict
             Parameter names mapped to their values.
         """
         out = dict()
@@ -534,13 +407,21 @@ class BaseEstimator(object):
 
     def set_params(self, **params):
         """Set the parameters of this estimator.
+
         The method works on simple estimators as well as on nested objects
         (such as pipelines). The latter have parameters of the form
         ``<component>__<parameter>`` so that it's possible to update each
         component of a nested object.
+
+        Parameters
+        ----------
+        **params : dict
+            Parameters.
+
         Returns
         -------
-        self
+        inst : instance
+            The object.
         """
         if not params:
             # Simple optimisation to gain speed (inspect is slow)
@@ -578,6 +459,9 @@ class BaseEstimator(object):
     # conditionals that are not satisfied by our objects (e.g.,
     # ``if type(self).__module__.startswith('sklearn.')``.
 
+    def _more_tags(self):
+        return _DEFAULT_TAGS
+
     def _get_tags(self):
         collected_tags = {}
         for base_class in reversed(inspect.getmro(self.__class__)):
@@ -719,6 +603,7 @@ class EmpiricalCovariance(BaseEstimator):
             is computed.
 
         """
+        from scipy import linalg
         # covariance = check_array(covariance)
         # set covariance
         self.covariance_ = covariance
@@ -737,6 +622,7 @@ class EmpiricalCovariance(BaseEstimator):
             The precision matrix associated to the current covariance object.
 
         """
+        from scipy import linalg
         if self.store_precision:
             precision = self.precision_
         else:
@@ -744,23 +630,21 @@ class EmpiricalCovariance(BaseEstimator):
         return precision
 
     def fit(self, X, y=None):
-        """Fits the Maximum Likelihood Estimator covariance model
-        according to the given training data and parameters.
+        """Fit the Maximum Likelihood Estimator covariance model.
 
         Parameters
         ----------
         X : array-like, shape = [n_samples, n_features]
           Training data, where n_samples is the number of samples and
           n_features is the number of features.
-
-        y : not used, present for API consistence purpose.
+        y : ndarray | None
+            Not used, present for API consistency.
 
         Returns
         -------
         self : object
             Returns self.
-
-        """
+        """  # noqa: E501
         # X = check_array(X)
         if self.assume_centered:
             self.location_ = np.zeros(X.shape[1])
@@ -773,8 +657,9 @@ class EmpiricalCovariance(BaseEstimator):
         return self
 
     def score(self, X_test, y=None):
-        """Computes the log-likelihood of a Gaussian data set with
-        `self.covariance_` as an estimator of its covariance matrix.
+        """Compute the log-likelihood of a Gaussian dataset.
+
+        Uses ``self.covariance_`` as an estimator of its covariance matrix.
 
         Parameters
         ----------
@@ -783,15 +668,14 @@ class EmpiricalCovariance(BaseEstimator):
             the number of samples and n_features is the number of features.
             X_test is assumed to be drawn from the same distribution than
             the data used in fit (including centering).
-
-        y : not used, present for API consistence purpose.
+        y : ndarray | None
+            Not used, present for API consistency.
 
         Returns
         -------
         res : float
             The likelihood of the data set with `self.covariance_` as an
             estimator of its covariance matrix.
-
         """
         # compute empirical covariance of the test set
         test_cov = empirical_covariance(
@@ -804,23 +688,19 @@ class EmpiricalCovariance(BaseEstimator):
     def error_norm(self, comp_cov, norm='frobenius', scaling=True,
                    squared=True):
         """Computes the Mean Squared Error between two covariance estimators.
-        (In the sense of the Frobenius norm).
 
         Parameters
         ----------
         comp_cov : array-like, shape = [n_features, n_features]
             The covariance to compare with.
-
         norm : str
             The type of norm used to compute the error. Available error types:
             - 'frobenius' (default): sqrt(tr(A^t.A))
             - 'spectral': sqrt(max(eigenvalues(A^t.A))
             where A is the error ``(comp_cov - self.covariance_)``.
-
         scaling : bool
             If True (default), the squared error norm is divided by n_features.
             If False, the squared error norm is not rescaled.
-
         squared : bool
             Whether to compute the squared error norm or the error norm.
             If True (default), the squared error norm is returned.
@@ -830,8 +710,8 @@ class EmpiricalCovariance(BaseEstimator):
         -------
         The Mean Squared Error (in the sense of the Frobenius norm) between
         `self` and `comp_cov` covariance estimators.
-
         """
+        from scipy import linalg
         # compute the error
         error = comp_cov - self.covariance_
         # compute the error norm
@@ -908,6 +788,7 @@ def log_likelihood(emp_cov, precision):
 
 def _logdet(A):
     """Compute the log det of a positive semidefinite matrix."""
+    from scipy import linalg
     vals = linalg.eigvalsh(A)
     # avoid negative (numerical errors) or zero (semi-definite matrix) values
     tol = vals.max() * vals.size * np.finfo(np.float64).eps
@@ -1096,12 +977,21 @@ def _crop_colorbar(cbar, cbar_vmin, cbar_vmax):
 
     # matplotlib >= 3.2.0 no longer normalizes axes between 0 and 1
     # See https://matplotlib.org/3.2.1/api/prev_api_changes/api_changes_3.2.0.html
+    # _outline was removed in
+    # https://github.com/matplotlib/matplotlib/commit/03a542e875eba091a027046d5ec652daa8be6863
+    # so we use the code from there
     if LooseVersion(matplotlib.__version__) >= LooseVersion("3.2.0"):
         cbar.ax.set_ylim(cbar_vmin, cbar_vmax)
         X, _ = cbar._mesh()
-        new_X = np.array([X[0], X[-1]])
-        new_Y = np.array([[cbar_vmin, cbar_vmin], [cbar_vmax, cbar_vmax]])
-        xy = cbar._outline(new_X, new_Y)
+        X = np.array([X[0], X[-1]])
+        Y = np.array([[cbar_vmin, cbar_vmin], [cbar_vmax, cbar_vmax]])
+        N = X.shape[0]
+        ii = [0, 1, N - 2, N - 1, 2 * N - 1, 2 * N - 2, N + 1, N, 0]
+        x = X.T.reshape(-1)[ii]
+        y = Y.T.reshape(-1)[ii]
+        xy = (np.column_stack([y, x])
+              if cbar.orientation == 'horizontal' else
+              np.column_stack([x, y]))
         cbar.outline.set_xy(xy)
     else:
         cbar.ax.set_ylim(cbar.norm(cbar_vmin), cbar.norm(cbar_vmax))
