@@ -11,14 +11,15 @@
 # License: Simplified BSD
 
 import numpy as np
-from scipy import sparse
 
 from .parametric import f_oneway, ttest_1samp_no_p
 from ..parallel import parallel_func, check_n_jobs
 from ..fixes import jit, has_numba
 from ..utils import (split_list, logger, verbose, ProgressBar, warn, _pl,
                      check_random_state, _check_option, _validate_type)
-from ..source_estimate import SourceEstimate
+from ..source_estimate import (SourceEstimate, VolSourceEstimate,
+                               MixedSourceEstimate)
+from ..source_space import SourceSpaces
 
 
 def _get_buddies_fallback(r, s, neighbors, indices=None):
@@ -110,16 +111,21 @@ def _masked_sum_power(x, c, t_power):
     return np.sum(np.sign(x[c]) * np.abs(x[c]) ** t_power)
 
 
+@jit()
+def _sum_cluster_data(data, tstep):
+    return np.sign(data) * np.logical_not(data == 0) * tstep
+
+
 def _get_clusters_spatial(s, neighbors):
     """Form spatial clusters using neighbor lists.
 
     This is equivalent to _get_components with n_times = 1, with a properly
-    reconfigured connectivity matrix (formed as "neighbors" list)
+    reconfigured adjacency matrix (formed as "neighbors" list)
     """
     # s is a vector of spatial indices that are significant, like:
     #     s = np.where(x_in)[0]
     # for x_in representing a single time-instant
-    r = np.ones(s.shape, np.bool)
+    r = np.ones(s.shape, bool)
     clusters = list()
     next_ind = 0 if s.size > 0 else -1
     while next_ind >= 0:
@@ -149,10 +155,10 @@ def _reassign(check, clusters, base, num):
 
 
 def _get_clusters_st_1step(keepers, neighbors):
-    """Directly calculate connectivity.
+    """Directly calculate clusters.
 
     This uses knowledge that time points are
-    only connected to adjacent neighbors for data organized as time x space.
+    only adjacent to immediate neighbors for data organized as time x space.
 
     This algorithm time increases linearly with the number of time points,
     compared to with the square for the standard (graph) algorithm.
@@ -197,10 +203,10 @@ def _get_clusters_st_1step(keepers, neighbors):
 
 
 def _get_clusters_st_multistep(keepers, neighbors, max_step=1):
-    """Directly calculate connectivity.
+    """Directly calculate clusters.
 
     This uses knowledge that time points are
-    only connected to adjacent neighbors for data organized as time x space.
+    only adjacent to immediate neighbors for data organized as time x space.
 
     This algorithm time increases linearly with the number of time points,
     compared to with the square for the standard (graph) algorithm.
@@ -252,16 +258,16 @@ def _get_clusters_st(x_in, neighbors, max_step=1):
     if len(cl_goods) > 0:
         keepers = [np.array([], dtype=int)] * n_times
         row, col = np.unravel_index(cl_goods, (n_times, n_src))
+        lims = [0]
         if isinstance(row, int):
             row = [row]
             col = [col]
-            lims = [0]
         else:
             order = np.argsort(row)
             row = row[order]
             col = col[order]
-            lims = [0] + (np.where(np.diff(row) > 0)[0] +
-                          1).tolist() + [len(row)]
+            lims += (np.where(np.diff(row) > 0)[0] + 1).tolist()
+            lims.append(len(row))
 
         for start, end in zip(lims[:-1], lims[1:]):
             keepers[row[start]] = np.sort(col[start:end])
@@ -274,23 +280,24 @@ def _get_clusters_st(x_in, neighbors, max_step=1):
         return []
 
 
-def _get_components(x_in, connectivity, return_list=True):
-    """Get connected components from a mask and a connectivity matrix."""
-    if connectivity is False:
+def _get_components(x_in, adjacency, return_list=True):
+    """Get connected components from a mask and a adjacency matrix."""
+    from scipy import sparse
+    if adjacency is False:
         components = np.arange(len(x_in))
     else:
         from scipy.sparse.csgraph import connected_components
-        mask = np.logical_and(x_in[connectivity.row], x_in[connectivity.col])
-        data = connectivity.data[mask]
-        row = connectivity.row[mask]
-        col = connectivity.col[mask]
-        shape = connectivity.shape
+        mask = np.logical_and(x_in[adjacency.row], x_in[adjacency.col])
+        data = adjacency.data[mask]
+        row = adjacency.row[mask]
+        col = adjacency.col[mask]
+        shape = adjacency.shape
         idx = np.where(x_in)[0]
         row = np.concatenate((row, idx))
         col = np.concatenate((col, idx))
         data = np.concatenate((data, np.ones(len(idx), dtype=data.dtype)))
-        connectivity = sparse.coo_matrix((data, (row, col)), shape=shape)
-        _, components = connected_components(connectivity)
+        adjacency = sparse.coo_matrix((data, (row, col)), shape=shape)
+        _, components = connected_components(adjacency)
     if return_list:
         start = np.min(components)
         stop = np.max(components)
@@ -305,7 +312,7 @@ def _get_components(x_in, connectivity, return_list=True):
         return components
 
 
-def _find_clusters(x, threshold, tail=0, connectivity=None, max_step=1,
+def _find_clusters(x, threshold, tail=0, adjacency=None, max_step=1,
                    include=None, partitions=None, t_power=1, show_info=False):
     """Find all clusters which are above/below a certain threshold.
 
@@ -322,17 +329,17 @@ def _find_clusters(x, threshold, tail=0, connectivity=None, max_step=1,
         threshold-free cluster enhancement.
     tail : -1 | 0 | 1
         Type of comparison
-    connectivity : scipy.sparse.coo_matrix, None, or list
-        Defines connectivity between features. The matrix is assumed to
+    adjacency : scipy.sparse.coo_matrix, None, or list
+        Defines adjacency between features. The matrix is assumed to
         be symmetric and only the upper triangular half is used.
-        If connectivity is a list, it is assumed that each entry stores the
+        If adjacency is a list, it is assumed that each entry stores the
         indices of the spatial neighbors in a spatio-temporal dataset x.
-        Default is None, i.e, a regular lattice connectivity.
-        False means no connectivity.
+        Default is None, i.e, a regular lattice adjacency.
+        False means no adjacency.
     max_step : int
-        If connectivity is a list, this defines the maximal number of steps
+        If adjacency is a list, this defines the maximal number of steps
         between vertices along the second dimension (typically time) to be
-        considered connected.
+        considered adjacent.
     include : 1D bool array or None
         Mask to apply to the data of points to cluster. If None, all points
         are used.
@@ -369,6 +376,10 @@ def _find_clusters(x, threshold, tail=0, connectivity=None, max_step=1,
             raise KeyError('threshold, if dict, must have at least '
                            '"start" and "step"')
         tfce = True
+        use_x = x[np.isfinite(x)]
+        if use_x.size == 0:
+            raise RuntimeError(
+                'No finite values found in the observed statistic values')
         if tail == -1:
             if threshold['start'] > 0:
                 raise ValueError('threshold["start"] must be <= 0 for '
@@ -376,11 +387,12 @@ def _find_clusters(x, threshold, tail=0, connectivity=None, max_step=1,
             if threshold['step'] >= 0:
                 raise ValueError('threshold["step"] must be < 0 for '
                                  'tail == -1')
-            stop = np.min(x)
+            stop = np.min(use_x)
         elif tail == 1:
-            stop = np.max(x)
+            stop = np.max(use_x)
         else:  # tail == 0
-            stop = np.max(np.abs(x))
+            stop = max(np.max(use_x), -np.min(use_x))
+        del use_x
         thresholds = np.arange(threshold['start'], stop,
                                threshold['step'], float)
         h_power = threshold.get('h_power', 2)
@@ -411,11 +423,10 @@ def _find_clusters(x, threshold, tail=0, connectivity=None, max_step=1,
 
     # set these here just in case thresholds == []
     clusters = list()
-    sums = np.empty(0)
+    sums = list()
     for ti, thresh in enumerate(thresholds):
         # these need to be reset on each run
         clusters = list()
-        sums = np.empty(0)
         if tail == 0:
             x_ins = [np.logical_and(x > thresh, include),
                      np.logical_and(x < -thresh, include)]
@@ -426,12 +437,12 @@ def _find_clusters(x, threshold, tail=0, connectivity=None, max_step=1,
         # loop over tails
         for x_in in x_ins:
             if np.any(x_in):
-                out = _find_clusters_1dir_parts(x, x_in, connectivity,
+                out = _find_clusters_1dir_parts(x, x_in, adjacency,
                                                 max_step, partitions, t_power,
                                                 ndimage)
                 clusters += out[0]
-                sums = np.concatenate((sums, out[1]))
-        if tfce is True:
+                sums.append(out[1])
+        if tfce:
             # the score of each point is the sum of the h^H * e^E for each
             # supporting section "rectangle" h x e.
             if ti == 0:
@@ -450,10 +461,12 @@ def _find_clusters(x, threshold, tail=0, connectivity=None, max_step=1,
                 else:
                     len_c = len(c)
                 scores[c] += h * (len_c ** e_power)
-    if tfce is True:
+    # turn sums into array
+    sums = np.concatenate(sums) if sums else np.array([])
+    if tfce:
         # each point gets treated independently
         clusters = np.arange(x.size)
-        if connectivity is None or connectivity is False:
+        if adjacency is None or adjacency is False:
             if x.ndim == 1:
                 # slices
                 clusters = [slice(c, c + 1) for c in clusters]
@@ -467,11 +480,11 @@ def _find_clusters(x, threshold, tail=0, connectivity=None, max_step=1,
     return clusters, sums
 
 
-def _find_clusters_1dir_parts(x, x_in, connectivity, max_step, partitions,
+def _find_clusters_1dir_parts(x, x_in, adjacency, max_step, partitions,
                               t_power, ndimage):
     """Deal with partitions, and pass the work to _find_clusters_1dir."""
     if partitions is None:
-        clusters, sums = _find_clusters_1dir(x, x_in, connectivity, max_step,
+        clusters, sums = _find_clusters_1dir(x, x_in, adjacency, max_step,
                                              t_power, ndimage)
     else:
         # cluster each partition separately
@@ -479,7 +492,7 @@ def _find_clusters_1dir_parts(x, x_in, connectivity, max_step, partitions,
         sums = list()
         for p in range(np.max(partitions) + 1):
             x_i = np.logical_and(x_in, partitions == p)
-            out = _find_clusters_1dir(x, x_i, connectivity, max_step, t_power,
+            out = _find_clusters_1dir(x, x_i, adjacency, max_step, t_power,
                                       ndimage)
             clusters += out[0]
             sums.append(out[1])
@@ -487,15 +500,17 @@ def _find_clusters_1dir_parts(x, x_in, connectivity, max_step, partitions,
     return clusters, sums
 
 
-def _find_clusters_1dir(x, x_in, connectivity, max_step, t_power, ndimage):
+def _find_clusters_1dir(x, x_in, adjacency, max_step, t_power, ndimage):
     """Actually call the clustering algorithm."""
-    if connectivity is None:
+    from scipy import sparse
+    if adjacency is None:
         labels, n_labels = ndimage.label(x_in)
 
         if x.ndim == 1:
             # slices
             clusters = ndimage.find_objects(labels, n_labels)
-            if len(clusters) == 0:
+            # equivalent to if len(clusters) == 0 but faster
+            if not clusters:
                 sums = list()
             else:
                 index = list(range(1, n_labels + 1))
@@ -509,24 +524,24 @@ def _find_clusters_1dir(x, x_in, connectivity, max_step, t_power, ndimage):
             # boolean masks (raveled)
             clusters = list()
             sums = np.empty(n_labels)
-            for l in range(1, n_labels + 1):
-                c = labels == l
+            for label in range(n_labels):
+                c = labels == label + 1
                 clusters.append(c.ravel())
                 if t_power == 1:
-                    sums[l - 1] = np.sum(x[c])
+                    sums[label] = np.sum(x[c])
                 else:
-                    sums[l - 1] = np.sum(np.sign(x[c]) *
+                    sums[label] = np.sum(np.sign(x[c]) *
                                          np.abs(x[c]) ** t_power)
     else:
         if x.ndim > 1:
-            raise Exception("Data should be 1D when using a connectivity "
+            raise Exception("Data should be 1D when using a adjacency "
                             "to define clusters.")
-        if isinstance(connectivity, sparse.spmatrix) or connectivity is False:
-            clusters = _get_components(x_in, connectivity)
-        elif isinstance(connectivity, list):  # use temporal adjacency
-            clusters = _get_clusters_st(x_in, connectivity, max_step)
+        if isinstance(adjacency, sparse.spmatrix) or adjacency is False:
+            clusters = _get_components(x_in, adjacency)
+        elif isinstance(adjacency, list):  # use temporal adjacency
+            clusters = _get_clusters_st(x_in, adjacency, max_step)
         else:
-            raise ValueError('Connectivity must be a sparse matrix or list')
+            raise ValueError('adjacency must be a sparse matrix or list')
         if t_power == 1:
             sums = [_masked_sum(x, c) for c in clusters]
         else:
@@ -543,11 +558,20 @@ def _cluster_indices_to_mask(components, n_tot):
     return components
 
 
-def _cluster_mask_to_indices(components):
+def _cluster_mask_to_indices(components, shape):
     """Convert to the old format of clusters, which were bool arrays."""
     for ci, c in enumerate(components):
-        if not isinstance(c, slice):
-            components[ci] = np.where(c)[0]
+        if isinstance(c, np.ndarray):  # mask
+            components[ci] = np.where(c.reshape(shape))
+        else:
+            assert isinstance(c, tuple), type(c)
+            c = list(c)  # tuple->list
+            for ii, cc in enumerate(c):
+                if isinstance(cc, slice):
+                    c[ii] = np.arange(cc.start, cc.stop)
+                else:
+                    c[ii] = np.where(cc)[0]
+            components[ci] = tuple(c)
     return components
 
 
@@ -557,8 +581,6 @@ def _pval_from_histogram(T, H0, tail):
     For each stat compute a p-value as percentile of its statistics
     within all statistics in surrogate data
     """
-    _check_option('tail', tail, [-1, 0, 1])
-
     # from pct to fraction
     if tail == -1:  # up tail
         pval = np.array([np.mean(H0 <= t) for t in T])
@@ -570,31 +592,32 @@ def _pval_from_histogram(T, H0, tail):
     return pval
 
 
-def _setup_connectivity(connectivity, n_tests, n_times):
-    if not sparse.issparse(connectivity):
-        raise ValueError("If connectivity matrix is given, it must be a "
+def _setup_adjacency(adjacency, n_tests, n_times):
+    from scipy import sparse
+    if not sparse.issparse(adjacency):
+        raise ValueError("If adjacency matrix is given, it must be a "
                          "SciPy sparse matrix.")
-    if connectivity.shape[0] == n_tests:  # use global algorithm
-        connectivity = connectivity.tocoo()
+    if adjacency.shape[0] == n_tests:  # use global algorithm
+        adjacency = adjacency.tocoo()
     else:  # use temporal adjacency algorithm
-        got_times, mod = divmod(n_tests, connectivity.shape[0])
+        got_times, mod = divmod(n_tests, adjacency.shape[0])
         if got_times != n_times or mod != 0:
             raise ValueError(
-                'connectivity (len %d) must be of the correct size, i.e. be '
+                'adjacency (len %d) must be of the correct size, i.e. be '
                 'equal to or evenly divide the number of tests (%d).\n\n'
-                'If connectivity was computed for a source space, try using '
+                'If adjacency was computed for a source space, try using '
                 'the fwd["src"] or inv["src"] as some original source space '
                 'vertices can be excluded during forward computation'
-                % (connectivity.shape[0], n_tests))
+                % (adjacency.shape[0], n_tests))
         # we claim to only use upper triangular part... not true here
-        connectivity = (connectivity + connectivity.transpose()).tocsr()
-        connectivity = [connectivity.indices[connectivity.indptr[i]:
-                                             connectivity.indptr[i + 1]]
-                        for i in range(len(connectivity.indptr) - 1)]
-    return connectivity
+        adjacency = (adjacency + adjacency.transpose()).tocsr()
+        adjacency = [
+            adjacency.indices[adjacency.indptr[i]:adjacency.indptr[i + 1]]
+            for i in range(len(adjacency.indptr) - 1)]
+    return adjacency
 
 
-def _do_permutations(X_full, slices, threshold, tail, connectivity, stat_fun,
+def _do_permutations(X_full, slices, threshold, tail, adjacency, stat_fun,
                      max_step, include, partitions, t_power, orders,
                      sample_shape, buffer_size, progress_bar):
     n_samp, n_vars = X_full.shape
@@ -636,13 +659,13 @@ def _do_permutations(X_full, slices, threshold, tail, connectivity, stat_fun,
                 tmp = stat_fun(*X_buffer)
                 t_obs_surr[pos: pos + n_var_loop] = tmp[:n_var_loop]
 
-        # The stat should have the same shape as the samples for no conn.
-        if connectivity is None:
+        # The stat should have the same shape as the samples for no adj.
+        if adjacency is None:
             t_obs_surr.shape = sample_shape
 
         # Find cluster on randomized stats
         out = _find_clusters(t_obs_surr, threshold=threshold, tail=tail,
-                             max_step=max_step, connectivity=connectivity,
+                             max_step=max_step, adjacency=adjacency,
                              partitions=partitions, include=include,
                              t_power=t_power)
         perm_clusters_sums = out[1]
@@ -657,7 +680,7 @@ def _do_permutations(X_full, slices, threshold, tail, connectivity, stat_fun,
     return max_cluster_sums
 
 
-def _do_1samp_permutations(X, slices, threshold, tail, connectivity, stat_fun,
+def _do_1samp_permutations(X, slices, threshold, tail, adjacency, stat_fun,
                            max_step, include, partitions, t_power, orders,
                            sample_shape, buffer_size, progress_bar):
     n_samp, n_vars = X.shape
@@ -706,13 +729,13 @@ def _do_1samp_permutations(X, slices, threshold, tail, connectivity, stat_fun,
                 tmp = stat_fun(X_flip_buffer)
                 t_obs_surr[pos: pos + n_var_loop] = tmp[:n_var_loop]
 
-        # The stat should have the same shape as the samples for no conn.
-        if connectivity is None:
+        # The stat should have the same shape as the samples for no adj.
+        if adjacency is None:
             t_obs_surr.shape = sample_shape
 
         # Find cluster on randomized stats
         out = _find_clusters(t_obs_surr, threshold=threshold, tail=tail,
-                             max_step=max_step, connectivity=connectivity,
+                             max_step=max_step, adjacency=adjacency,
                              partitions=partitions, include=include,
                              t_power=t_power)
         perm_clusters_sums = out[1]
@@ -806,7 +829,7 @@ def _get_1samp_orders(n_samples, n_permutations, tail, rng):
 
 
 def _permutation_cluster_test(X, threshold, n_permutations, tail, stat_fun,
-                              connectivity, n_jobs, seed, max_step,
+                              adjacency, n_jobs, seed, max_step,
                               exclude, step_down_p, t_power, out_type,
                               check_disjoint, buffer_size):
     n_jobs = check_n_jobs(n_jobs)
@@ -817,6 +840,7 @@ def _permutation_cluster_test(X, threshold, n_permutations, tail, stat_fun,
     is elicited.
     """
     _check_option('out_type', out_type, ['mask', 'indices'])
+    _check_option('tail', tail, [-1, 0, 1])
     if not isinstance(threshold, dict):
         threshold = float(threshold)
         if (tail < 0 and threshold > 0 or tail > 0 and threshold < 0 or
@@ -838,8 +862,8 @@ def _permutation_cluster_test(X, threshold, n_permutations, tail, stat_fun,
     X = [np.reshape(x, (x.shape[0], -1)) for x in X]
     n_tests = X[0].shape[1]
 
-    if connectivity is not None and connectivity is not False:
-        connectivity = _setup_connectivity(connectivity, n_tests, n_times)
+    if adjacency is not None and adjacency is not False:
+        adjacency = _setup_adjacency(adjacency, n_tests, n_times)
 
     if (exclude is not None) and not exclude.size == n_tests:
         raise ValueError('exclude must be the same shape as X[0]')
@@ -862,12 +886,12 @@ def _permutation_cluster_test(X, threshold, n_permutations, tail, stat_fun,
                  'Setting buffer_size to None.')
             buffer_size = None
 
-    # The stat should have the same shape as the samples for no conn.
+    # The stat should have the same shape as the samples for no adj.
     if t_obs.size != np.prod(sample_shape):
         raise ValueError('t_obs.shape %s provided by stat_fun %s is not '
                          'compatible with the sample shape %s'
                          % (t_obs.shape, stat_fun, sample_shape))
-    if connectivity is None or connectivity is False:
+    if adjacency is None or adjacency is False:
         t_obs.shape = sample_shape
 
     if exclude is not None:
@@ -875,36 +899,37 @@ def _permutation_cluster_test(X, threshold, n_permutations, tail, stat_fun,
     else:
         include = None
 
-    # determine if connectivity itself can be separated into disjoint sets
-    if check_disjoint is True and (connectivity is not None and
-                                   connectivity is not False):
-        partitions = _get_partitions_from_connectivity(connectivity, n_times)
+    # determine if adjacency itself can be separated into disjoint sets
+    if check_disjoint is True and (adjacency is not None and
+                                   adjacency is not False):
+        partitions = _get_partitions_from_adjacency(adjacency, n_times)
     else:
         partitions = None
     logger.info('Running initial clustering')
-    out = _find_clusters(t_obs, threshold, tail, connectivity,
+    out = _find_clusters(t_obs, threshold, tail, adjacency,
                          max_step=max_step, include=include,
                          partitions=partitions, t_power=t_power,
                          show_info=True)
     clusters, cluster_stats = out
+
+    # The stat should have the same shape as the samples
+    t_obs.shape = sample_shape
+
     # For TFCE, return the "adjusted" statistic instead of raw scores
     if isinstance(threshold, dict):
-        t_obs = cluster_stats.copy()
+        t_obs = cluster_stats.reshape(t_obs.shape) * np.sign(t_obs)
 
     logger.info('Found %d clusters' % len(clusters))
 
     # convert clusters to old format
-    if connectivity is not None and connectivity is not False:
+    if adjacency is not None and adjacency is not False:
         # our algorithms output lists of indices by default
         if out_type == 'mask':
             clusters = _cluster_indices_to_mask(clusters, n_tests)
     else:
         # ndimage outputs slices or boolean masks by default
         if out_type == 'indices':
-            clusters = _cluster_mask_to_indices(clusters)
-
-    # The stat should have the same shape as the samples
-    t_obs.shape = sample_shape
+            clusters = _cluster_mask_to_indices(clusters, t_obs.shape)
 
     # convert our seed to orders
     # check to see if we can do an exact test
@@ -956,7 +981,7 @@ def _permutation_cluster_test(X, threshold, n_permutations, tail, stat_fun,
         logger.info('Permuting %d times%s...' % (len(orders), extra))
         with ProgressBar(len(orders)) as progress_bar:
             H0 = parallel(
-                my_do_perm_func(X_full, slices, threshold, tail, connectivity,
+                my_do_perm_func(X_full, slices, threshold, tail, adjacency,
                                 stat_fun, max_step, this_include, partitions,
                                 t_power, order, sample_shape, buffer_size,
                                 progress_bar.subset(idx))
@@ -980,7 +1005,7 @@ def _permutation_cluster_test(X, threshold, n_permutations, tail, stat_fun,
         step_down_include = np.ones(n_tests, dtype=bool)
         for ti in to_remove:
             step_down_include[clusters[ti]] = False
-        if connectivity is None and connectivity is not False:
+        if adjacency is None and adjacency is not False:
             step_down_include.shape = sample_shape
         n_step_downs += 1
         if step_down_p > 0:
@@ -1032,17 +1057,18 @@ def _check_fun(X, stat_fun, threshold, tail=0, kind='within'):
 @verbose
 def permutation_cluster_test(
         X, threshold=None, n_permutations=1024, tail=0, stat_fun=None,
-        connectivity=None, n_jobs=1, seed=None, max_step=1, exclude=None,
-        step_down_p=0, t_power=1, out_type='mask', check_disjoint=False,
+        adjacency=None, n_jobs=1, seed=None, max_step=1, exclude=None,
+        step_down_p=0, t_power=1, out_type='indices', check_disjoint=False,
         buffer_size=1000, verbose=None):
     """Cluster-level statistical permutation test.
 
-    For a list of nd-arrays of data, e.g. 2d for time series or 3d for
-    time-frequency power values, calculate some statistics corrected for
-    multiple comparisons using permutations and cluster level correction.
-    Each element of the list X contains the data for one group of
-    observations. Randomized data are generated with random partitions
-    of the data. See :footcite:`MarisOostenveld2007` for details.
+    For a list of :class:`NumPy arrays <numpy.ndarray>` of data,
+    calculate some statistics corrected for multiple comparisons using
+    permutations and cluster-level correction. Each element of the list ``X``
+    should contain the data for one group of observations (e.g., 2D arrays for
+    time series, 3D arrays for time-frequency power values). Permutations are
+    generated with random partitions of the data. See
+    :footcite:`MarisOostenveld2007` for details.
 
     Parameters
     ----------
@@ -1053,12 +1079,16 @@ def permutation_cluster_test(
         the size of a single observation. For example if ``X = [X1, X2]``
         with ``X1.shape = (20, 50, 4)`` and ``X2.shape = (17, 50, 4)``, then
         ``X`` has 2 groups with respectively 20 and 17 observations in each,
-        and each data point is of shape ``(50, 4)``.
+        and each data point is of shape ``(50, 4)``. Note: that the
+        *last dimension* of each element of ``X`` should correspond to the
+        dimension represented in the ``adjacency`` parameter
+        (e.g., spectral data should be provided as
+        ``(observations, frequencies, channels/vertices)``).
     %(clust_thresh_f)s
     %(clust_nperm_int)s
     %(clust_tail)s
     %(clust_stat_f)s
-    %(clust_con_n)s
+    %(clust_adj_n)s
     %(n_jobs)s
     %(seed)s
     %(clust_maxstep)s
@@ -1069,7 +1099,6 @@ def permutation_cluster_test(
     %(clust_stepdown)s
     %(clust_power_f)s
     %(clust_out)s
-        Default is ``'mask'``.
     %(clust_disjoint)s
     %(clust_buffer)s
     %(verbose)s
@@ -1092,7 +1121,7 @@ def permutation_cluster_test(
     stat_fun, threshold = _check_fun(X, stat_fun, threshold, tail, 'between')
     return _permutation_cluster_test(
         X=X, threshold=threshold, n_permutations=n_permutations, tail=tail,
-        stat_fun=stat_fun, connectivity=connectivity, n_jobs=n_jobs, seed=seed,
+        stat_fun=stat_fun, adjacency=adjacency, n_jobs=n_jobs, seed=seed,
         max_step=max_step, exclude=exclude, step_down_p=step_down_p,
         t_power=t_power, out_type=out_type, check_disjoint=check_disjoint,
         buffer_size=buffer_size)
@@ -1101,9 +1130,9 @@ def permutation_cluster_test(
 @verbose
 def permutation_cluster_1samp_test(
         X, threshold=None, n_permutations=1024, tail=0, stat_fun=None,
-        connectivity=None, verbose=None, n_jobs=1, seed=None, max_step=1,
-        exclude=None, step_down_p=0, t_power=1, out_type='mask',
-        check_disjoint=False, buffer_size=1000):
+        adjacency=None, n_jobs=1, seed=None, max_step=1,
+        exclude=None, step_down_p=0, t_power=1, out_type='indices',
+        check_disjoint=False, buffer_size=1000, verbose=None):
     """Non-parametric cluster-level paired t-test.
 
     Parameters
@@ -1119,8 +1148,7 @@ def permutation_cluster_1samp_test(
     %(clust_nperm_all)s
     %(clust_tail)s
     %(clust_stat_t)s
-    %(clust_con_1)s
-    %(verbose)s
+    %(clust_adj_1)s
     %(n_jobs)s
     %(seed)s
     %(clust_maxstep)s
@@ -1131,7 +1159,6 @@ def permutation_cluster_1samp_test(
     %(clust_stepdown)s
     %(clust_power_t)s
     %(clust_out)s
-        Default is ``'mask'``.
     %(clust_disjoint)s
     %(clust_buffer)s
     %(verbose)s
@@ -1178,7 +1205,7 @@ def permutation_cluster_1samp_test(
     stat_fun, threshold = _check_fun(X, stat_fun, threshold, tail)
     return _permutation_cluster_test(
         X=[X], threshold=threshold, n_permutations=n_permutations, tail=tail,
-        stat_fun=stat_fun, connectivity=connectivity, n_jobs=n_jobs, seed=seed,
+        stat_fun=stat_fun, adjacency=adjacency, n_jobs=n_jobs, seed=seed,
         max_step=max_step, exclude=exclude, step_down_p=step_down_p,
         t_power=t_power, out_type=out_type, check_disjoint=check_disjoint,
         buffer_size=buffer_size)
@@ -1187,7 +1214,7 @@ def permutation_cluster_1samp_test(
 @verbose
 def spatio_temporal_cluster_1samp_test(
         X, threshold=None, n_permutations=1024, tail=0,
-        stat_fun=None, connectivity=None, n_jobs=1, seed=None,
+        stat_fun=None, adjacency=None, n_jobs=1, seed=None,
         max_step=1, spatial_exclude=None, step_down_p=0, t_power=1,
         out_type='indices', check_disjoint=False, buffer_size=1000,
         verbose=None):
@@ -1207,7 +1234,7 @@ def spatio_temporal_cluster_1samp_test(
     %(clust_nperm_all)s
     %(clust_tail)s
     %(clust_stat_t)s
-    %(clust_con_st1)s
+    %(clust_adj_st1)s
     %(n_jobs)s
     %(seed)s
     %(clust_maxstep)s
@@ -1216,11 +1243,9 @@ def spatio_temporal_cluster_1samp_test(
     %(clust_stepdown)s
     %(clust_power_t)s
     %(clust_out)s
-        Default is ``'indices'``.
     %(clust_disjoint)s
     %(clust_buffer)s
     %(verbose)s
-    %(verbose)s
 
     Returns
     -------
@@ -1246,7 +1271,7 @@ def spatio_temporal_cluster_1samp_test(
         exclude = None
     return permutation_cluster_1samp_test(
         X, threshold=threshold, stat_fun=stat_fun, tail=tail,
-        n_permutations=n_permutations, connectivity=connectivity,
+        n_permutations=n_permutations, adjacency=adjacency,
         n_jobs=n_jobs, seed=seed, max_step=max_step, exclude=exclude,
         step_down_p=step_down_p, t_power=t_power, out_type=out_type,
         check_disjoint=check_disjoint, buffer_size=buffer_size)
@@ -1255,9 +1280,10 @@ def spatio_temporal_cluster_1samp_test(
 @verbose
 def spatio_temporal_cluster_test(
         X, threshold=None, n_permutations=1024, tail=0, stat_fun=None,
-        connectivity=None, verbose=None, n_jobs=1, seed=None, max_step=1,
+        adjacency=None, n_jobs=1, seed=None, max_step=1,
         spatial_exclude=None, step_down_p=0, t_power=1, out_type='indices',
-        check_disjoint=False, buffer_size=1000):
+        check_disjoint=False, buffer_size=1000,
+        verbose=None):
     """Non-parametric cluster-level test for spatio-temporal data.
 
     This function provides a convenient wrapper for
@@ -1277,8 +1303,7 @@ def spatio_temporal_cluster_test(
     %(clust_nperm_int)s
     %(clust_tail)s
     %(clust_stat_f)s
-    %(clust_con_stn)s
-    %(verbose)s
+    %(clust_adj_stn)s
     %(n_jobs)s
     %(seed)s
     %(clust_maxstep)s
@@ -1287,9 +1312,9 @@ def spatio_temporal_cluster_test(
     %(clust_stepdown)s
     %(clust_power_f)s
     %(clust_out)s
-        Default is ``'indices'``.
     %(clust_disjoint)s
     %(clust_buffer)s
+    %(verbose)s
 
     Returns
     -------
@@ -1315,14 +1340,14 @@ def spatio_temporal_cluster_test(
         exclude = None
     return permutation_cluster_test(
         X, threshold=threshold, stat_fun=stat_fun, tail=tail,
-        n_permutations=n_permutations, connectivity=connectivity,
+        n_permutations=n_permutations, adjacency=adjacency,
         n_jobs=n_jobs, seed=seed, max_step=max_step, exclude=exclude,
         step_down_p=step_down_p, t_power=t_power, out_type=out_type,
         check_disjoint=check_disjoint, buffer_size=buffer_size)
 
 
 def _st_mask_from_s_inds(n_times, n_vertices, vertices, set_as=True):
-    """Compute mask to apply to a spatio-temporal connectivity matrix.
+    """Compute mask to apply to a spatio-temporal adjacency matrix.
 
     This can be used to include (or exclude) certain spatial coordinates.
     This is useful for excluding certain regions from analysis (e.g.,
@@ -1354,30 +1379,30 @@ def _st_mask_from_s_inds(n_times, n_vertices, vertices, set_as=True):
 
 
 @verbose
-def _get_partitions_from_connectivity(connectivity, n_times, verbose=None):
-    """Specify disjoint subsets (e.g., hemispheres) based on connectivity."""
-    if isinstance(connectivity, list):
-        test = np.ones(len(connectivity))
-        test_conn = np.zeros((len(connectivity), len(connectivity)),
-                             dtype='bool')
-        for vi in range(len(connectivity)):
-            test_conn[connectivity[vi], vi] = True
-        test_conn = sparse.coo_matrix(test_conn, dtype='float')
+def _get_partitions_from_adjacency(adjacency, n_times, verbose=None):
+    """Specify disjoint subsets (e.g., hemispheres) based on adjacency."""
+    from scipy import sparse
+    if isinstance(adjacency, list):
+        test = np.ones(len(adjacency))
+        test_adj = np.zeros((len(adjacency), len(adjacency)), dtype='bool')
+        for vi in range(len(adjacency)):
+            test_adj[adjacency[vi], vi] = True
+        test_adj = sparse.coo_matrix(test_adj, dtype='float')
     else:
-        test = np.ones(connectivity.shape[0])
-        test_conn = connectivity
+        test = np.ones(adjacency.shape[0])
+        test_adj = adjacency
 
-    part_clusts = _find_clusters(test, 0, 1, test_conn)[0]
+    part_clusts = _find_clusters(test, 0, 1, test_adj)[0]
     if len(part_clusts) > 1:
-        logger.info('%i disjoint connectivity sets found'
+        logger.info('%i disjoint adjacency sets found'
                     % len(part_clusts))
         partitions = np.zeros(len(test), dtype='int')
         for ii, pc in enumerate(part_clusts):
             partitions[pc] = ii
-        if isinstance(connectivity, list):
+        if isinstance(adjacency, list):
             partitions = np.tile(partitions, n_times)
     else:
-        logger.info('No disjoint connectivity sets found')
+        logger.info('No disjoint adjacency sets found')
         partitions = None
 
     return partitions
@@ -1416,9 +1441,13 @@ def summarize_clusters_stc(clu, p_thresh=0.05, tstep=1.0, tmin=0,
         The time of the first sample.
     subject : str
         The name of the subject.
-    vertices : list of array | None
+    vertices : list of array | instance of SourceSpaces | None
         The vertex numbers associated with the source space locations. Defaults
-        to None. If None, equals ```[np.arange(10242), np.arange(10242)]```.
+        to None. If None, equals ``[np.arange(10242), np.arange(10242)]``.
+        Can also be an instance of SourceSpaces to get vertex numbers from.
+
+        .. versionchanged:: 0.21
+           Added support for SourceSpaces.
 
     Returns
     -------
@@ -1428,12 +1457,30 @@ def summarize_clusters_stc(clu, p_thresh=0.05, tstep=1.0, tmin=0,
         contain each individual cluster. The magnitude of the activity
         corresponds to the duration spanned by the cluster (duration units are
         determined by ``tstep``).
+
+        .. versionchanged:: 0.21
+           Added support for volume and mixed source estimates.
     """
+    _validate_type(vertices, (None, list, SourceSpaces), 'vertices')
     if vertices is None:
         vertices = [np.arange(10242), np.arange(10242)]
+        klass = SourceEstimate
+    elif isinstance(vertices, SourceSpaces):
+        klass = dict(surface=SourceEstimate,
+                     volume=VolSourceEstimate,
+                     mixed=MixedSourceEstimate)[vertices.kind]
+        vertices = [s['vertno'] for s in vertices]
+    else:
+        klass = {1: VolSourceEstimate,
+                 2: SourceEstimate}.get(len(vertices), MixedSourceEstimate)
+    n_vertices_need = sum(len(v) for v in vertices)
 
     t_obs, clusters, clu_pvals, _ = clu
     n_times, n_vertices = t_obs.shape
+    if n_vertices != n_vertices_need:
+        raise ValueError(
+            f'Number of cluster vertices ({n_vertices}) did not match the '
+            f'provided vertices ({n_vertices_need})')
     good_cluster_inds = np.where(clu_pvals < p_thresh)[0]
 
     #  Build a convenient representation of each cluster, where each
@@ -1446,15 +1493,13 @@ def summarize_clusters_stc(clu, p_thresh=0.05, tstep=1.0, tmin=0,
     data_summary = np.zeros((n_vertices, len(good_cluster_inds) + 1))
     for ii, cluster_ind in enumerate(good_cluster_inds):
         data.fill(0)
-        v_inds = clusters[cluster_ind][1]
-        t_inds = clusters[cluster_ind][0]
+        t_inds, v_inds = clusters[cluster_ind]
         data[v_inds, t_inds] = t_obs[t_inds, v_inds]
         # Store a nice visualization of the cluster by summing across time
-        data = np.sign(data) * np.logical_not(data == 0) * tstep
-        data_summary[:, ii + 1] = np.sum(data, axis=1)
+        data_summary[:, ii + 1] = np.sum(_sum_cluster_data(data, tstep),
+                                         axis=1)
         # Make the first "time point" a sum across all clusters for easy
         # visualization
     data_summary[:, 0] = np.sum(data_summary, axis=1)
 
-    return SourceEstimate(data_summary, vertices, tmin=tmin, tstep=tstep,
-                          subject=subject)
+    return klass(data_summary, vertices, tmin, tstep, subject)
