@@ -21,7 +21,7 @@ from mne.utils import (_get_inst_data, hashfunc,
                        _undo_scaling_array, _PCA, requires_sklearn,
                        _array_equal_nan, _julian_to_cal, _cal_to_julian,
                        _dt_to_julian, _julian_to_dt, grand_average,
-                       _ReuseCycle, requires_version)
+                       _ReuseCycle, requires_version, numerics)
 
 
 base_dir = op.join(op.dirname(__file__), '..', '..', 'io', 'tests', 'data')
@@ -304,6 +304,21 @@ def test_object_size():
         size = object_size(obj)
         assert lower < size < upper, \
             '%s < %s < %s:\n%s' % (lower, size, upper, obj)
+    # views work properly
+    x = dict(a=1)
+    assert object_size(x) < 1000
+    x['a'] = np.ones(100000, float)
+    nb = x['a'].nbytes
+    sz = object_size(x)
+    assert nb < sz < nb * 1.01
+    x['b'] = x['a']
+    sz = object_size(x)
+    assert nb < sz < nb * 1.01
+    x['b'] = x['a'].view()
+    x['b'].flags.writeable = False
+    assert x['a'].flags.writeable
+    sz = object_size(x)
+    assert nb < sz < nb * 1.01
 
 
 def test_object_diff_with_nan():
@@ -405,7 +420,7 @@ def test_hash():
 
 
 @requires_sklearn
-@pytest.mark.parametrize('n_components', (None, 0.8, 8, 'mle'))
+@pytest.mark.parametrize('n_components', (None, 0.9999, 8, 'mle'))
 @pytest.mark.parametrize('whiten', (True, False))
 def test_pca(n_components, whiten):
     """Test PCA equivalence."""
@@ -421,12 +436,13 @@ def test_pca(n_components, whiten):
     X_mne = pca_mne.fit_transform(X)
     assert_array_equal(X, X_orig)
     assert_allclose(X_skl, X_mne)
+    assert pca_mne.n_components_ == pca_skl.n_components_
     for key in ('mean_', 'components_',
                 'explained_variance_', 'explained_variance_ratio_'):
         val_skl, val_mne = getattr(pca_skl, key), getattr(pca_mne, key)
         assert_allclose(val_skl, val_mne)
     if isinstance(n_components, float):
-        assert 1 < pca_mne.n_components_ < n_dim
+        assert pca_mne.n_components_ == n_dim - 1
     elif isinstance(n_components, int):
         assert pca_mne.n_components_ == n_components
     elif n_components == 'mle':
@@ -514,3 +530,12 @@ def test_reuse_cycle():
         iterable.restore('a')
     assert ''.join(next(iterable) for _ in range(4)) == 'acde'
     assert ''.join(next(iterable) for _ in range(5)) == 'abcde'
+
+
+@pytest.mark.parametrize('n', (0, 1, 10, 1000))
+@pytest.mark.parametrize('d', (0.0001, 1, 2.5, 1000))
+def test_arange_div(numba_conditional, n, d):
+    """Test Numba arange_div."""
+    want = np.arange(n) / d
+    got = numerics._arange_div(n, d)
+    assert_allclose(got, want)
