@@ -7,7 +7,6 @@
 #
 # License: Simplified BSD
 
-from mne.minimum_norm.inverse import apply_inverse
 import os.path as op
 from pathlib import Path
 import sys
@@ -29,15 +28,15 @@ from mne.io import (read_raw_ctf, read_raw_bti, read_raw_kit, read_info,
 from mne.io._digitization import write_dig
 from mne.io.pick import pick_info
 from mne.io.constants import FIFF
+from mne.minimum_norm import apply_inverse
 from mne.viz import (plot_sparse_source_estimates, plot_source_estimates,
                      snapshot_brain_montage, plot_head_positions,
                      plot_alignment, plot_sensors_connectivity,
                      plot_brain_colorbar, link_brains, mne_analyze_colormap)
 from mne.viz._3d import _process_clim, _linearize_map, _get_map_ticks
 from mne.viz.utils import _fake_click
-from mne.utils import (requires_pysurfer, run_tests_if_main,
-                       requires_nibabel, traits_test, catch_logging,
-                       run_subprocess, modified_env)
+from mne.utils import (requires_nibabel, traits_test,
+                       catch_logging, run_subprocess, modified_env)
 from mne.datasets import testing
 from mne.source_space import read_source_spaces
 from mne.bem import read_bem_solution, read_bem_surfaces
@@ -98,14 +97,12 @@ def test_plot_head_positions():
     pytest.raises(ValueError, plot_head_positions, pos, 'foo')
     with pytest.raises(ValueError, match='shape'):
         plot_head_positions(pos, axes=1.)
-    plt.close('all')
 
 
 @testing.requires_testing_data
-@requires_pysurfer
 @traits_test
 @pytest.mark.slowtest
-def test_plot_sparse_source_estimates(renderer_interactive):
+def test_plot_sparse_source_estimates(renderer_interactive, brain_gc):
     """Test plotting of (sparse) source estimates."""
     sample_src = read_source_spaces(src_fname)
 
@@ -121,13 +118,15 @@ def test_plot_sparse_source_estimates(renderer_interactive):
     stc = SourceEstimate(stc_data, vertices, 1, 1)
 
     colormap = 'mne_analyze'
-    plot_source_estimates(stc, 'sample', colormap=colormap,
-                          background=(1, 1, 0),
-                          subjects_dir=subjects_dir, colorbar=True,
-                          clim='auto')
-    pytest.raises(TypeError, plot_source_estimates, stc, 'sample',
-                  figure='foo', hemi='both', clim='auto',
-                  subjects_dir=subjects_dir)
+    brain = plot_source_estimates(
+        stc, 'sample', colormap=colormap, background=(1, 1, 0),
+        subjects_dir=subjects_dir, colorbar=True, clim='auto')
+    brain.close()
+    del brain
+    with pytest.raises(TypeError, match='figure must be'):
+        plot_source_estimates(
+            stc, 'sample', figure='foo', hemi='both', clim='auto',
+            subjects_dir=subjects_dir)
 
     # now do sparse version
     vertices = sample_src[0]['vertno']
@@ -166,7 +165,7 @@ def test_plot_evoked_field(renderer):
 @pytest.mark.slowtest  # can be slow on OSX
 @testing.requires_testing_data
 @traits_test
-def test_plot_alignment(tmpdir, renderer):
+def test_plot_alignment(tmpdir, renderer, mixed_fwd_cov_evoked):
     """Test plotting of -trans.fif files and MEG sensor layouts."""
     # generate fiducials file for testing
     tempdir = str(tmpdir)
@@ -210,6 +209,13 @@ def test_plot_alignment(tmpdir, renderer):
                   src=sample_src)
     sample_src.plot(subjects_dir=subjects_dir, head=True, skull=True,
                     brain='white')
+    # mixed source space
+    mixed_src = mixed_fwd_cov_evoked[0]['src']
+    assert mixed_src.kind == 'mixed'
+    plot_alignment(info, meg=['helmet', 'sensors'], dig=True,
+                   coord_frame='head', trans=Path(trans_fname),
+                   subject='sample', mri_fiducials=fiducials_path,
+                   subjects_dir=subjects_dir, src=mixed_src)
     renderer.backend._close_all()
     # no-head version
     renderer.backend._close_all()
@@ -281,15 +287,19 @@ def test_plot_alignment(tmpdir, renderer):
                    src=src, dig=True, surfaces=['brain', 'inner_skull',
                                                 'outer_skull', 'outer_skin'])
     sphere = make_sphere_model('auto', None, evoked.info)  # one layer
+    # if you ask for a brain surface with a 1-layer sphere model it's an error
+    with pytest.raises(RuntimeError, match='Sphere model does not have'):
+        fig = plot_alignment(subject='sample', subjects_dir=subjects_dir,
+                             surfaces=['brain'], bem=sphere)
+    # but you can ask for a specific brain surface, and
     # no info is permitted
     fig = plot_alignment(trans=trans_fname, subject='sample', meg=False,
                          coord_frame='mri', subjects_dir=subjects_dir,
-                         surfaces=['brain'], bem=sphere, show_axes=True)
+                         surfaces=['white'], bem=sphere, show_axes=True)
     renderer.backend._close_all()
     if renderer._get_3d_backend() == 'mayavi':
         import mayavi  # noqa: F401 analysis:ignore
         assert isinstance(fig, mayavi.core.scene.Scene)
-
     # 3D coil with no defined draw (ConvexHull)
     info_cube = pick_info(info, [0])
     info['dig'] = None
@@ -303,7 +313,7 @@ def test_plot_alignment(tmpdir, renderer):
         plot_alignment(info_cube, meg='sensors', surfaces=(), dig=True)
 
     # one layer bem with skull surfaces:
-    with pytest.raises(ValueError, match='sphere conductor model must have'):
+    with pytest.raises(RuntimeError, match='Sphere model does not.*boundary'):
         plot_alignment(info=info, trans=trans_fname,
                        subject='sample', subjects_dir=subjects_dir,
                        surfaces=['brain', 'head', 'inner_skull'], bem=sphere)
@@ -320,7 +330,7 @@ def test_plot_alignment(tmpdir, renderer):
         plot_alignment(info=info, trans=trans_fname,
                        subject='sample', subjects_dir=subjects_dir,
                        surfaces=['white', 'pial'])
-    with pytest.raises(TypeError, match='all entries in surfaces must be'):
+    with pytest.raises(TypeError, match='surfaces.*must be'):
         plot_alignment(info=info, trans=trans_fname,
                        subject='sample', subjects_dir=subjects_dir,
                        surfaces=[1])
@@ -328,6 +338,14 @@ def test_plot_alignment(tmpdir, renderer):
         plot_alignment(info=info, trans=trans_fname,
                        subject='sample', subjects_dir=subjects_dir,
                        surfaces=['foo'])
+    with pytest.raises(TypeError, match="must be an instance of "):
+        plot_alignment(info=info, trans=trans_fname,
+                       subject='sample', subjects_dir=subjects_dir,
+                       surfaces=dict(brain='super clear'))
+    with pytest.raises(ValueError, match="must be between 0 and 1"):
+        plot_alignment(info=info, trans=trans_fname,
+                       subject='sample', subjects_dir=subjects_dir,
+                       surfaces=dict(brain=42))
     fwd_fname = op.join(data_dir, 'MEG', 'sample',
                         'sample_audvis_trunc-meg-eeg-oct-4-fwd.fif')
     fwd = read_forward_solution(fwd_fname)
@@ -338,7 +356,10 @@ def test_plot_alignment(tmpdir, renderer):
     plot_alignment(subject='sample', subjects_dir=subjects_dir,
                    trans=trans_fname, fwd=fwd,
                    surfaces='white', coord_frame='head')
-
+    # surfaces as dict
+    plot_alignment(subject='sample', coord_frame='head',
+                   subjects_dir=subjects_dir,
+                   surfaces={'white': 0.4, 'outer_skull': 0.6, 'head': None})
     # fNIRS (default is pairs)
     info = read_raw_nirx(nirx_fname).info
     with catch_logging() as log:
@@ -363,12 +384,12 @@ def test_plot_alignment(tmpdir, renderer):
 
 @pytest.mark.slowtest  # can be slow on OSX
 @testing.requires_testing_data
-@requires_pysurfer
 @traits_test
-def test_process_clim_plot(renderer_interactive):
+def test_process_clim_plot(renderer_interactive, brain_gc):
     """Test functionality for determining control points with stc.plot."""
     sample_src = read_source_spaces(src_fname)
-    kwargs = dict(subjects_dir=subjects_dir, smoothing_steps=1)
+    kwargs = dict(subjects_dir=subjects_dir, smoothing_steps=1,
+                  time_viewer=False, show_traces=False)
 
     vertices = [s['vertno'] for s in sample_src]
     n_time = 5
@@ -378,19 +399,28 @@ def test_process_clim_plot(renderer_interactive):
     stc = SourceEstimate(stc_data, vertices, 1, 1, 'sample')
 
     # Test for simple use cases
-    stc.plot(**kwargs)
-    stc.plot(clim=dict(pos_lims=(10, 50, 90)), **kwargs)
-    stc.plot(colormap='hot', clim='auto', **kwargs)
-    stc.plot(colormap='mne', clim='auto', **kwargs)
-    stc.plot(clim=dict(kind='value', lims=(10, 50, 90)), figure=99, **kwargs)
-    pytest.raises(TypeError, stc.plot, clim='auto', figure=[0], **kwargs)
+    brain = stc.plot(**kwargs)
+    assert brain.data['center'] is None
+    brain.close()
+    brain = stc.plot(clim=dict(pos_lims=(10, 50, 90)), **kwargs)
+    assert brain.data['center'] == 0.
+    brain.close()
+    brain = stc.plot(colormap='hot', clim='auto', **kwargs)
+    brain.close()
+    brain = stc.plot(colormap='mne', clim='auto', **kwargs)
+    brain.close()
+    brain = stc.plot(clim=dict(kind='value', lims=(10, 50, 90)), figure=99,
+                     **kwargs)
+    brain.close()
+    with pytest.raises(TypeError, match='must be a'):
+        stc.plot(clim='auto', figure=[0], **kwargs)
 
     # Test for correct clim values
     with pytest.raises(ValueError, match='monotonically'):
         stc.plot(clim=dict(kind='value', pos_lims=[0, 1, 0]), **kwargs)
     with pytest.raises(ValueError, match=r'.*must be \(3,\)'):
         stc.plot(colormap='mne', clim=dict(pos_lims=(5, 10, 15, 20)), **kwargs)
-    with pytest.raises(ValueError, match="'value', 'values' and 'percent'"):
+    with pytest.raises(ValueError, match="'value', 'values', and 'percent'"):
         stc.plot(clim=dict(pos_lims=(5, 10, 15), kind='foo'), **kwargs)
     with pytest.raises(ValueError, match='must be "auto" or dict'):
         stc.plot(colormap='mne', clim='foo', **kwargs)
@@ -405,7 +435,8 @@ def test_process_clim_plot(renderer_interactive):
     # Test handling of degenerate data: thresholded maps
     stc._data.fill(0.)
     with pytest.warns(RuntimeWarning, match='All data were zero'):
-        plot_source_estimates(stc, **kwargs)
+        brain = plot_source_estimates(stc, **kwargs)
+    brain.close()
 
 
 def _assert_mapdata_equal(a, b):
@@ -485,24 +516,22 @@ def test_stc_mpl():
     stc_data = np.ones((n_verts * n_time))
     stc_data.shape = (n_verts, n_time)
     stc = SourceEstimate(stc_data, vertices, 1, 1, 'sample')
-    with pytest.warns(RuntimeWarning, match='not included'):
-        stc.plot(subjects_dir=subjects_dir, time_unit='s', views='ven',
-                 hemi='rh', smoothing_steps=2, subject='sample',
-                 backend='matplotlib', spacing='oct1', initial_time=0.001,
-                 colormap='Reds')
-        fig = stc.plot(subjects_dir=subjects_dir, time_unit='ms', views='dor',
-                       hemi='lh', smoothing_steps=2, subject='sample',
-                       backend='matplotlib', spacing='ico2', time_viewer=True,
-                       colormap='mne')
-        time_viewer = fig.time_viewer
-        _fake_click(time_viewer, time_viewer.axes[0], (0.5, 0.5))  # change t
-        time_viewer.canvas.key_press_event('ctrl+right')
-        time_viewer.canvas.key_press_event('left')
+    stc.plot(subjects_dir=subjects_dir, time_unit='s', views='ven',
+             hemi='rh', smoothing_steps=7, subject='sample',
+             backend='matplotlib', spacing='oct1', initial_time=0.001,
+             colormap='Reds')
+    fig = stc.plot(subjects_dir=subjects_dir, time_unit='ms', views='dor',
+                   hemi='lh', smoothing_steps=7, subject='sample',
+                   backend='matplotlib', spacing='ico2', time_viewer=True,
+                   colormap='mne')
+    time_viewer = fig.time_viewer
+    _fake_click(time_viewer, time_viewer.axes[0], (0.5, 0.5))  # change t
+    time_viewer.canvas.key_press_event('ctrl+right')
+    time_viewer.canvas.key_press_event('left')
     pytest.raises(ValueError, stc.plot, subjects_dir=subjects_dir,
                   hemi='both', subject='sample', backend='matplotlib')
     pytest.raises(ValueError, stc.plot, subjects_dir=subjects_dir,
                   time_unit='ss', subject='sample', backend='matplotlib')
-    plt.close('all')
 
 
 @pytest.mark.timeout(60)  # can sometimes take > 60 sec
@@ -525,10 +554,9 @@ def test_plot_dipole_mri_orthoview(coord_frame, idx, show_all, title):
     fig.canvas.key_press_event('up')
     fig.canvas.key_press_event('down')
     fig.canvas.key_press_event('a')  # some other key
-    ax = plt.subplot(111)
-    pytest.raises(TypeError, dipoles.plot_locations, trans, 'sample',
-                  subjects_dir, ax=ax)
-    plt.close('all')
+    ax = fig.add_subplot(211)
+    with pytest.raises(TypeError, match='instance of Axes3D'):
+        dipoles.plot_locations(trans, 'sample', subjects_dir, ax=ax)
 
 
 @testing.requires_testing_data
@@ -569,16 +597,15 @@ def test_snapshot_brain_montage(renderer):
 
 @pytest.mark.slowtest  # can be slow on OSX
 @testing.requires_testing_data
-@requires_pysurfer
-@traits_test
 @pytest.mark.parametrize('pick_ori', ('vector', None))
 @pytest.mark.parametrize('kind', ('surface', 'volume', 'mixed'))
 def test_plot_source_estimates(renderer_interactive, all_src_types_inv_evoked,
-                               pick_ori, kind):
+                               pick_ori, kind, brain_gc):
     """Test plotting of scalar and vector source estimates."""
+    backend = renderer_interactive._get_3d_backend()
+    is_pyvista = backend != 'mayavi'
     invs, evoked = all_src_types_inv_evoked
     inv = invs[kind]
-    is_pyvista = renderer_interactive._get_3d_backend() == 'pyvista'
     with pytest.warns(None):  # PCA mag
         stc = apply_inverse(evoked, inv, pick_ori=pick_ori)
     stc.data[1] *= -1  # make it signed
@@ -592,7 +619,7 @@ def test_plot_source_estimates(renderer_interactive, all_src_types_inv_evoked,
                   )
     if pick_ori != 'vector':
         kwargs['surface'] = 'white'
-        kwargs['backend'] = renderer_interactive._get_3d_backend()
+        kwargs['backend'] = backend
     # Mayavi can't handle non-surface
     if kind != 'surface' and not is_pyvista:
         with pytest.raises(RuntimeError, match='PyVista'):
@@ -602,6 +629,11 @@ def test_plot_source_estimates(renderer_interactive, all_src_types_inv_evoked,
     brain.close()
     del brain
 
+    these_kwargs = kwargs.copy()
+    these_kwargs['show_traces'] = 'foo'
+    with pytest.raises(ValueError, match='show_traces'):
+        meth(**these_kwargs)
+    del these_kwargs
     if pick_ori == 'vector':
         with pytest.raises(ValueError, match='use "pos_lims"'):
             meth(**kwargs, clim=dict(pos_lims=[1, 2, 3]))
@@ -621,10 +653,14 @@ def test_plot_source_estimates(renderer_interactive, all_src_types_inv_evoked,
 
     # flatmaps (mostly a lot of error checking)
     these_kwargs = kwargs.copy()
-    these_kwargs.update(surface='flat', views='auto')
+    these_kwargs.update(surface='flat', views='auto', hemi='both',
+                        verbose='debug')
     if kind == 'surface' and pick_ori != 'vector' and is_pyvista:
-        with pytest.raises(FileNotFoundError, match='flatmap'):
-            meth(**these_kwargs)  # sample does not have them
+        with catch_logging() as log:
+            with pytest.raises(FileNotFoundError, match='flatmap'):
+                meth(**these_kwargs)  # sample does not have them
+        log = log.getvalue()
+        assert 'offset: 0' in log
     fs_stc = stc.copy()
     fs_stc.subject = 'fsaverage'  # this is wrong, but don't have to care
     flat_meth = getattr(fs_stc, meth_key)
@@ -640,6 +676,7 @@ def test_plot_source_estimates(renderer_interactive, all_src_types_inv_evoked,
     else:
         brain = flat_meth(**these_kwargs)
         brain.close()
+        del brain
         these_kwargs.update(surface='inflated', views='flat')
         with pytest.raises(ValueError, match='surface="flat".*views="flat"'):
             flat_meth(**these_kwargs)
@@ -684,14 +721,26 @@ def test_plot_sensors_connectivity(renderer):
     n_channels = len(picks)
     con = np.random.RandomState(42).randn(n_channels, n_channels)
     info = raw.info
-    with pytest.raises(TypeError):
-        plot_sensors_connectivity(info='foo', con=con,
-                                  picks=picks)
-    with pytest.raises(ValueError):
-        plot_sensors_connectivity(info=info, con=con[::2, ::2],
-                                  picks=picks)
+    with pytest.raises(TypeError, match='must be an instance of Info'):
+        plot_sensors_connectivity(info='foo', con=con, picks=picks)
+    with pytest.raises(ValueError, match='does not correspond to the size'):
+        plot_sensors_connectivity(info=info, con=con[::2, ::2], picks=picks)
 
-    plot_sensors_connectivity(info=info, con=con, picks=picks)
+    fig = plot_sensors_connectivity(info=info, con=con, picks=picks)
+    if renderer._get_3d_backend() == 'pyvista':
+        try:
+            # pyvista<0.30.0
+            title = fig.plotter.scalar_bar.GetTitle()
+        except AttributeError:
+            # pyvista>=0.30.0
+            title = list(fig.plotter.scalar_bars.values())[0].GetTitle()
+    else:
+        assert renderer._get_3d_backend() == 'mayavi'
+        # the last thing we add is the Tube, so we need to go
+        # vtkDataSource->Stripper->Tube->ModuleManager
+        mod_man = fig.children[-1].children[0].children[0].children[0]
+        title = mod_man.scalar_lut_manager.scalar_bar.title
+    assert title == 'Connectivity'
 
 
 @pytest.mark.parametrize('orientation', ('horizontal', 'vertical'))
@@ -718,20 +767,16 @@ def test_brain_colorbar(orientation, diverging, lims):
     else:
         ticks = lims
     plt.draw()
-    # old mpl always spans 0->1 for the actual ticks, so we need to
-    # look at the labels
     assert_array_equal(
         [float(h.get_text().replace('−', '-')) for h in have()], ticks)
     assert_array_equal(empty(), [])
-    plt.close('all')
 
 
 @pytest.mark.slowtest  # slow-ish on Travis OSX
-@requires_pysurfer
 @testing.requires_testing_data
 @traits_test
 def test_mixed_sources_plot_surface(renderer_interactive):
-    """Test plot_surface() for  mixed source space."""
+    """Test plot_surface() for mixed source space."""
     src = read_source_spaces(fwd_fname2)
     N = np.sum([s['nuse'] for s in src])  # number of sources
 
@@ -744,9 +789,11 @@ def test_mixed_sources_plot_surface(renderer_interactive):
 
     stc = MixedSourceEstimate(data, vertno, 0, 1)
 
-    stc.surface().plot(views='lat', hemi='split',
-                       subject='fsaverage', subjects_dir=subjects_dir,
-                       colorbar=False)
+    brain = stc.surface().plot(views='lat', hemi='split',
+                               subject='fsaverage', subjects_dir=subjects_dir,
+                               colorbar=False)
+    brain.close()
+    del brain
 
 
 @testing.requires_testing_data
@@ -772,7 +819,7 @@ def test_link_brains(renderer_interactive):
         subjects_dir=subjects_dir, colorbar=True,
         clim='auto'
     )
-    if renderer_interactive._get_3d_backend() != 'pyvista':
+    if renderer_interactive._get_3d_backend() == 'mayavi':
         with pytest.raises(NotImplementedError, match='backend is pyvista'):
             link_brains(brain)
     else:
@@ -792,6 +839,3 @@ def test_renderer(renderer):
            'assert backend == %r, backend' % (backend,)]
     with modified_env(MNE_3D_BACKEND=backend):
         run_subprocess(cmd)
-
-
-run_tests_if_main()
